﻿using System;
using System.Linq;
using IdeaBlade.EntityModel;
using IdeaBlade.Extensions;
using Model;
using ValidationLab;

// ReSharper disable LocalizableElement
// ReSharper disable LocalizableElement

namespace ConsoleApp
{
    class Program
    {
        static void Main()
        {
            new Program().Run();
        }

        public void Run()
        {
            Manager = new NorthwindIBEntities();
            CanAccessDatabase(); // sanity check

            GetAlfreds();
            GetSimons();

            Console.WriteLine("\n+++ Alfreds Customer Unique Query Tests +++");
            Manager.RemoveEntity(Simons); // remove Simon from cache

            AlfredsCanBeNameOfEntityWithAlfredsId();
            AlfredsCannotBeRenamedToSimons();
            AlfredsCannotBeNameOfDifferentEntityInCache();

            Console.WriteLine("\n+++ Simons Customer Unique Verifier Tests +++");
            Manager.AttachEntity(Simons);  // restore Simons to the cache
            Manager.RemoveEntity(Alfreds); // remove Alfreds from cache

            SimonsNameCanBeChangedToSimon();
            SimonsNameCannotBeChangedToAlfreds();
            SimonsCannotBeNameOfDifferentEntityInCache();

            // Testing OrderDetail because it has a multi-part key and want to ensure that
            // IsDuplicateQueryBuilder handles multi-part keys properly
            Console.WriteLine("\n+++ OrderDetail Multi-part Key Unique Verifier Tests +++");
            OrderDetailCannotSetQuantity10();

            DevForceResourceManagerEmulatorTests();

            HoldConsoleOpen();
        }

        #region Alfreds Customer Uniqueness Query Tests

        // Alfreds is one of the NorthwindIB customers
        // These tests validate the query for duplicates, apart from the verifier.

        private void AlfredsCanBeNameOfEntityWithAlfredsId()
        {
            // The reported value should be TRUE
            Console.WriteLine("Unique test returned '{0}' when Alfreds name is for Alfreds entity.",
                              IsCustomerNameUnique(Alfreds, Alfreds.CompanyName));
        }

        private void AlfredsCannotBeRenamedToSimons()
        {
            // The reported value should be FALSE
            Console.WriteLine("Unique test returned '{0}' when Alfreds new name is for Simons entity in DATABASE.",
                              IsCustomerNameUnique(Alfreds, SimonsName));
        }

        private void AlfredsCannotBeNameOfDifferentEntityInCache()
        {
            // Add a new Customer to cache with Alfreds name
            // This sets up a "duplicate name" conflict
            var conflictCustomer = new Customer
                                       {
                                           CompanyName = Alfreds.CompanyName,
                                           CustomerID = Guid.NewGuid()
                                       };
            Manager.AddEntity(conflictCustomer);

            // The name of the "real" Alfreds Customer is no longer unique
            // because we created a new entity with its name and added that to cache
            // The reported value should be FALSE
            Console.WriteLine("Unique test returned '{0}' when Alfreds name is for non-Alfreds entity in CACHE.",
                              IsCustomerNameUnique(Alfreds, Alfreds.CompanyName));
            // Note that if we didn't check the cache, the reported value would be TRUE ... BAD BAD BAD!
        }

        #endregion

        #region Simons Customer Unique Verifier Tests

        // Simons is one of the NorthwindIB customers
        // These tests validate the UniqueVerifier

        // A Unique attribute was added to Customer.CompanyName in Customer.cs

        private void SimonsNameCanBeChangedToSimon()
        {
            Simons.CompanyName = "Some crazy name";
            if (SimonsHasErrors)
            {
                Console.WriteLine("BAD! Simons.CompanyName has errors when it set to 'crazy'. First error is '{0}'",
                                  SimonsFirstErrorMessage);
            }
            Simons.CompanyName = SimonsName; // the name in the db
            if (SimonsHasErrors)
            {
                Console.WriteLine("BAD! Simons.CompanyName has errors when its name is restored. First error is '{0}'",
                                  SimonsFirstErrorMessage);
            }
            else
            {
                Console.WriteLine("GOOD! Simons name can be set to its database value.");
            }
        }


        private void SimonsNameCannotBeChangedToAlfreds()
        {
            Simons.CompanyName = Alfreds.CompanyName; // a name in the db
            if (SimonsHasErrors)
            {
                Console.Write("GOOD! Simons.CompanyName has expected validation error: '{0}'.",
                                  SimonsFirstErrorMessage);

                Console.WriteLine(" Simons pending CompanyName should be unchanged; it is '{0}'.",
                    Simons.CompanyName);
            }
            else
            {
                Console.WriteLine("BAD! Simons name can be set Alfreds w/o validation error; it is '{0}'.",
                    Simons.CompanyName);
            }
        }

        private void SimonsCannotBeNameOfDifferentEntityInCache()
        {
            const string newCustomerName = "<new customer>";

            // Add a new Customer to cache 
            var conflictCustomer = new Customer
            {
                CompanyName = newCustomerName,
                CustomerID = Guid.NewGuid()
            };
            Manager.AddEntity(conflictCustomer);

            Simons.CompanyName = newCustomerName; // the name of entity in cache
            if (SimonsHasErrors)
            {
                Console.Write("GOOD! Simons has expected validation error: '{0}'.",
                                  SimonsFirstErrorMessage);

                Console.WriteLine(" Simons' pending CompanyName should be unchanged; it is '{0}'.",
                    Simons.CompanyName);
            }
            else
            {
                Console.WriteLine("BAD! Simons's name can be set to a name of an entity in cache; it is '{0}'.",
                    Simons.CompanyName);
                // Note that we'd be here if we didn't check the cache ... BAD BAD BAD!
            }
        }
        #endregion

        #region OrderDetail Unique Tests

        // Testing OrderDetail because it has a multi-part key and want to ensure that
        // IsDuplicateQueryBuilder handles multi-part keys properly

        // A Unique attribute was added to OrderDetail.Quantity in OrderDetail.cs

        private void OrderDetailCannotSetQuantity10()
        {
            var orderDetailWithQty12 =
                Manager.OrderDetails.Where(od => od.Quantity == 12).First();

            orderDetailWithQty12.Quantity = 10; // Known to occur frequently in db
            if (orderDetailWithQty12.EntityAspect.ValidationErrors.Any())
            {
                Console.Write("GOOD! orderDetailWithQty12 has expected validation error: '{0}'.",
                                  orderDetailWithQty12.EntityAspect.ValidationErrors.First().Message);

                Console.WriteLine(" orderDetailWithQty12 pending quantity should be unchanged; it is '{0}'.",
                    orderDetailWithQty12.Quantity);
            }
            else
            {
                Console.WriteLine("BAD! Can set quantity of orderDetailWithQty12 to 10 w/o validation error.");
            }
        }

        #endregion
 
        #region DevForceResourceManagerEmulatorTests

        private void DevForceResourceManagerEmulatorTests()
        {
            Console.WriteLine("\n+++ DevForceResourceManagerEmulator tests");

            // Set the custom ResourceManager for localization of validation error messages
            new IdeaBlade.Validation.VerifierEngine().ErrorsResourceManager = CustomResources.ResourceManager;

            Console.WriteLine("'VerifierMaxStringLength' emulator resource lookup is '{0}'.",
                              DevForceResourceManagerEmulator.Instance.GetString("VerifierMaxStringLength"));

            Console.WriteLine("'nonsense' emulator resource lookup should be the empty string, '{0}'.",
                              DevForceResourceManagerEmulator.Instance.GetString("nonsense"));

            Console.WriteLine("'Test' emulator resource lookup is '{0}'.",
                              DevForceResourceManagerEmulator.Instance.GetString("Test"));

            Console.WriteLine("'Customer_ContactName' emulator resource lookup is '{0}'.",
                              DevForceResourceManagerEmulator.Instance.GetString("Customer_ContactName"));

            ContactNameUsesCustomValidationMessage();
            RequiredValidationMessageIsReplaced();
        }

        private void ContactNameUsesCustomValidationMessage()
        {
            var aCustomer = new Customer {CustomerID = Guid.NewGuid(), CompanyName = "Acme"};
            Manager.AddEntity(aCustomer);
            aCustomer.ContactName = "A very very very very very very long contact name";

            var valResult = aCustomer.EntityAspect.ValidationErrors.FirstOrDefault();
            Console.WriteLine(
                (null == valResult)
                    ? "BAD! Expected string length validation error"
                    : String.Format("GOOD! Validation error with custom message: '{0}'.", valResult.Message)
                );
        }

        private void RequiredValidationMessageIsReplaced()
        {
            var aCustomer = new Customer { CustomerID = Guid.NewGuid(), CompanyName = "Acme" };
            Manager.AddEntity(aCustomer);
            aCustomer.CompanyName = null;

            var valResult = aCustomer.EntityAspect.ValidationErrors.FirstOrDefault();
            Console.WriteLine(
                (null == valResult)
                    ? "BAD! Expected required validation error"
                    : String.Format("GOOD! Validation error shows replaced 'VerifierRequired' custom message: '{0}'.", valResult.Message)
                );
        }
        #endregion

        #region IsCustomerNameUnique

        /// <summary>
        /// For smoke testing <see cref="IsDuplicateQueryBuilder"/>.
        /// If the proposed CompanyName is in use for a Customer with a different Id
        /// </summary>
        /// <param name="cust">Customer which would receive this name.</param>
        /// <param name="proposedName">Proposed Customer.CompanyName.</param>
        /// <returns>
        /// True if a Customer entity with this Id can receive this name; 
        /// false if a different Customer entity already has this name.
        /// </returns>
        public bool IsCustomerNameUnique(Customer cust, string proposedName)
        {
            var isDuplicate = IsDuplicateQueryBuilder.Build(cust, "CompanyName", proposedName);

            // Check the cache first
            if (isDuplicate
                .With(QueryStrategy.CacheOnly)
                .Any()
                )
                // Break on "return false" to see when duplicate is in cache
                return false;


            // No duplicate in cache; check the database
            return !isDuplicate
                        .With(QueryStrategy.DataSourceOnly)
                        .Any();
        }

        #endregion

        #region Helpers

        public NorthwindIBEntities Manager { get; set; }

        private void CanAccessDatabase()
        {
            // Getting the count proves we can access the database
            Console.WriteLine("Customers: {0}", Manager.Customers.Count());
        }

        private static void HoldConsoleOpen()
        {
            Console.WriteLine("\n***Press any key to end");
            Console.ReadKey();
        }

        public void GetAlfreds()
        {
            Alfreds = Manager.Customers.First(c => c.CustomerID == AlfredsId);
        }

        private void GetSimons()
        {
            Simons = Manager.Customers.First(c => c.CustomerID == SimonsId);
            SimonsName = Simons.CompanyName;
        }

        private bool SimonsHasErrors
        {
            get { return Simons.EntityAspect.ValidationErrors.Any(); }
        }

        private string SimonsFirstErrorMessage
        {
            get { return Simons.EntityAspect.ValidationErrors.First().Message; }
        }

        public Customer Alfreds { get; private set; }
        public Customer Simons { get; private set; }
        public string SimonsName { get; set; }
        public static Guid AlfredsId = new Guid("785efa04-cbf2-4dd7-a7de-083ee17b6ad2");
        public static Guid SimonsId = new Guid("657d32bc-158c-4d11-ac49-18981ce1e205");

        #endregion

    }

}
