using System;
using System.Threading;

namespace IdeaBlade.TestFramework {

  /// <summary>
  /// Derived WorkItems may choose to ask a <see cref="SynchronizationContext"/>
  /// to execute their actions.
  /// See <see cref="SyncContext"/>.
  /// </summary>
  public abstract class SynchronizedWorkItem : WorkItem {

    #region SynchronizationContext related

    /// <summary>
    /// Get and set <see cref="SynchronizationContext"/> to be used
    /// for <see cref="Send(Action)"/> and <see cref="Post(Action)"/> calls.
    /// </summary>
    /// <remarks>
    /// Get guarantees that such a SyncContext exists; if it can't find one
    /// it creates one ... which simply executes the action on the current thread.
    /// </remarks>
    protected SynchronizationContext SyncContext {
      get {
        return _synchronizationContext ??
               (_synchronizationContext = SynchronizationContext.Current ?? new SynchronizationContext());
      }
      set { _synchronizationContext = value; }
    }

    private SynchronizationContext _synchronizationContext;

    /// <summary>
    /// Send the action to the <see cref="SyncContext"/> for execution.
    /// See also <see cref="SynchronizationContext.Send"/>.
    /// </summary>
    /// <remarks>Uses a thread-local counter for the state object.</remarks>
    protected void Send(Action act) {
      Send(act, _stateCounter++);
    }

    /// <summary>
    /// Send the action and state object to the <see cref="SyncContext"/> for execution.
    /// See also <see cref="SynchronizationContext.Send"/>.
    /// </summary>
    protected void Send(Action act, object state) {
      Send(s => act(), state);
    }

    /// <summary>
    /// Send the action and state object to the <see cref="SyncContext"/> for execution.
    /// See also <see cref="SynchronizationContext.Send"/>.
    /// </summary>
    protected void Send(SendOrPostCallback act, object state) {
      SyncContext.Send(act, state);
    }

    /// <summary>
    /// Post the action to the <see cref="SyncContext"/> for execution.
    /// See also <see cref="SynchronizationContext.Post"/>.
    /// </summary>
    /// <remarks>Uses a thread-local counter for the state object.</remarks>
    protected void Post(Action act) {
      Post(act, _stateCounter++);
    }

    /// <summary>
    /// Post the action and state object to the <see cref="SyncContext"/> for execution.
    /// See also <see cref="SynchronizationContext.Post"/>.
    /// </summary>
    protected void Post(Action act, object state) {
      Post(s => act(), state);
    }
    /// <summary>
    /// Post the action and state object to the <see cref="SyncContext"/> for execution.
    /// See also <see cref="SynchronizationContext.Post"/>.
    /// </summary>
    protected void Post(SendOrPostCallback act, object state) {
      SyncContext.Post(act, state);
    }

    [ThreadStatic]
    private static int _stateCounter;

    #endregion

  }
}