using System;
using System.Reflection;
using IdeaBlade.EntityModel;

namespace IdeaBlade.TestFramework {

  public partial class DevForceTest {

    /// <summary>
    /// Try synchronous act and then complete (end) the test.
    /// Catch exception and route it to test harness.
    /// </summary>
    protected void TryToComplete(Action act) {
      Try(act);
      WaitFor.TestComplete(); 
    }

    /// <summary>
    /// Try synchronous act.
    /// Catch exception and route it to test harness.</summary>
    protected void Try(Action act) {
      WaitFor.It(act);
    }

    protected readonly NestedWaitFor WaitFor;

    /// <summary>
    /// An attempt at more intuitive synonyms for 
    /// Silverlight Testing Framework's Enqueue methods.
    /// Also affords consistency with WaitFor extension methods,
    /// <see cref="OperationTestForErrorExtensions"/>.
    /// </summary>
    /// <remarks>
    /// A coding experiment. Not sure this is such a great idea
    /// Perhaps more clever than good.
    /// </remarks>
    protected class NestedWaitFor {

      public NestedWaitFor(DevForceTest devForceTest) {
        DevForceTest = devForceTest;
      }

      protected DevForceTest DevForceTest { get; private set; }

      /// <summary>
      /// Add a Callback method into the test task queue. 
      /// Synonym for the <see cref="Callback"/> method.
      /// </summary>
      /// <param name="testCallback">
      /// Void-returning delegate, anonymous delegates work fine too.
      /// </param>
      public void It(Action testCallback) {
        DevForceTest.WaitForIt(testCallback);
      }

      /// <summary>
      /// Add a Callback method into the test task queue. 
      /// Similar to the PumpMessages(...) call, with the difference being 
      /// that there is no longer a single requirement: 
      /// you can enqueue several callback methods and other test tasks, 
      /// all of which will execute before the test completes and/or the engine continues.
      /// </summary>
      /// <param name="testCallback">
      /// Void-returning delegate, anonymous delegates work fine too.
      /// </param>
      public void Callback(Action testCallback) {
        DevForceTest.EnqueueCallback(testCallback);
      }

      public void Delay(double milliseconds) {
        DevForceTest.EnqueueDelay(milliseconds);
      }
      /// <summary>
      /// Delay a minimum amount of time before continuing. Similar to a sleep call that is not a blocking call.
      /// </summary>
      /// <param name="delay">
      /// The minimum time span to wait before continuing
      /// </param>
      public void Delay(TimeSpan delay) {
        DevForceTest.EnqueueDelay(delay);
      }
      // Based on Justin Angel: http://blogs.silverlight.net/blogs/justinangel/archive/2009/02/25/silverlight-unit-testing-rhinomocks-unity-and-resharper.aspx
      public T Event<T>(T objectToWaitForItsEvent, string eventName) {

        EventInfo eventInfo = objectToWaitForItsEvent.GetType().GetEvent(eventName);

        bool eventRaised = false;

        if (typeof(System.Windows.RoutedEventHandler).IsAssignableFrom(eventInfo.EventHandlerType))
          eventInfo.AddEventHandler(objectToWaitForItsEvent, (System.Windows.RoutedEventHandler) delegate { eventRaised = true; });

        else if (typeof(EventHandler).IsAssignableFrom(eventInfo.EventHandlerType))
          eventInfo.AddEventHandler(objectToWaitForItsEvent, (EventHandler) delegate { eventRaised = true; });

        DevForceTest.EnqueueConditional(() => eventRaised);

        return objectToWaitForItsEvent;
      }

      /// <summary>
      /// Wait until the DevForce <see cref="BaseOperation.IsCompleted"/> flag is true.
      /// Uses the <see cref="Conditional"/> method.
      /// </summary>
      /// <param name="op">Operation object returned from a DevForce async method.</param>
      public void IsCompleted(BaseOperation op) {
        IsCompleted(() => op);
      }

      /// <summary>
      /// Wait until the DevForce <see cref="BaseOperation.IsCompleted"/> flag is true.
      /// Uses the <see cref="Conditional"/> method.
      /// </summary>
      /// <param name="opFn">Fnc rtning operation object returned from a DevForce async method.</param>
      public void IsCompleted(Func<BaseOperation> opFn) {
        DevForceTest.EnqueueConditional(() => {
          var op = opFn();
          return op.IsCompleted || op.HasError || op.Cancelled;
        });
      }

      /// <summary>
      /// Instructs the test task queue to wait until the conditional call returns True 
      /// to continue executing other test tasks and/or ending the test method.
      /// </summary>
      /// <param name="conditional">
      /// Conditional method or delegate. Test will halt until this condition returns True.
      /// </param>
      public void Conditional(Func<bool> conditional) {
        DevForceTest.EnqueueConditional(conditional);
      }

      /// <summary>
      /// Enqueue a test task which calls the <see cref="TestComplete"/> method.
      /// </summary>
      public void TestComplete() {
        DevForceTest.EnqueueTestComplete();
      }
      /// <summary>
      /// Wait for <see cref="TestComplete"/> signal that 
      /// queued async tasks are done.
      /// See <see cref="WaitForTasksToFinish"/>.
      /// </summary>
      public void TasksToFinish() {
        DevForceTest.WaitForTasksToFinish();
      }
    }
  }
}