using System;
using System.Collections.ObjectModel;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace IdeaBlade.TestFramework {

  /// <summary>
  /// Base class for Microsoft Unit Test classes with asynchronous calls.
  /// Provides a uniform approach to asynchronous testing in both
  /// Full ("Desktop") .NET and Silverlight environments.
  /// See its most important method, <see cref="DoItAsync"/>,
  /// for an example of writing an asynch Test Method.
  /// </summary>
  public partial class DevForceTest {

    public DevForceTest() {
      WaitFor = new NestedWaitFor(this);
    }

    [ThreadStatic]
    public static DevForceTest TestClassBase;

    [TestInitialize]
    public void TestInitialize() {
      SetTestClassBase();
      Context();
    }

    /// <summary>
    /// Override to provide specific context for test methods of derived class.
    /// Extends this base class' MS Test [TestInitialize] method.
    /// </summary>
    protected virtual void Context() { }

    /// <summary>
    /// Perform the async action and block the test method thread
    /// until the action finishes.
    /// </summary>
    /// <param name="act">
    /// Void-returning delegate, anonymous delegates work fine too.
    /// </param>
    /// <remarks>
    /// Wrap an asynchronous action and its tests so that the Test Runner
    /// will hold the test method open until the async action completes.
    /// <para>
    /// Combines EnqueueCallback(Action) with WaitForTasksToFinish.
    /// </para>
    /// </remarks>
    /// <example>
    /// In this test which gets all customers asynchronously 
    /// the query is wrapped in a DoItAsync(...) call.
    /// <para>
    /// The parameter is an action that performs the query
    /// asynchronously and calls "TestComplete()" when done.
    /// </para><para>
    /// TestComplete tells the Test Runner that this test method
    /// has actually finished. Essential to call this method
    /// or else test Mmthod hangs indefinitely.
    /// It's a good idea to decorate your test method with 
    /// a <see cref="TimeoutAttribute"/> 
    /// to ensure that test method eventually closes.
    /// </para><para>
    /// The Test Runner automatically closes the test when the action throws
    /// an exception ... as it does when an assertion fails.
    /// </para>
    /// <![CDATA[
    /// [TestMethod, Asynchronous, Timeout(20000)]
    /// public void then_can_get_customers() {
    ///    DoItAsync(() =>
    /// 
    ///      Manager
    ///        .Customers
    ///        .ExecuteAsync()
    ///        .Completed += (sender, args) => {
    ///           args.HasError.ShouldBeFalse("Query failed with exception: " + args.Error); 
    ///           args.Results.ShouldNotBeEmpty("Should have customers"); 
    ///           TestComplete(); // <--- don't forget this!
    ///         }
    ///    );
    /// }
    /// ]]>
    /// </example>
    protected internal void DoItAsync(Action act) {
      EnqueueCallback(act);
      WaitForTasksToFinish();
    }

    /// <summary>
    /// Add a Callback method into the test task queue. 
    /// Synonym for the EnqueueCallback(Action) method.
    /// </summary>
    /// <param name="testCallback">
    /// Void-returning delegate, anonymous delegates work fine too.
    /// </param>
    protected internal void WaitForIt(Action testCallback) {
      EnqueueCallback(testCallback);
    }

    partial void SetTestClassBase();

  }
}