using IdeaBlade.EntityModel;
using IdeaBlade.Linq;

namespace Model
{
    /// <summary>
    /// Builder of queries to test for duplicate entity property values.
    /// </summary>
    public static class IsDuplicateQueryBuilder
    {
        /// <summary>
        /// Build a query to test for duplicate entity property values.
        /// </summary>
        /// <remarks>
        /// Produces a query that returns entities with matching proposed property value.
        /// excluding the <see cref="entity"/> submitted as a parameter to the function.
        /// The returned entities are considered "duplicates" of that (excluded) entity.
        /// For example, to get a query returning Customers other than 'someCustomer' 
        /// which have CompanyName == "Acme", try: 
        /// IsDuplicateQueryBuilder.Build(someCustomer, "CompanyName", "Acme")
        /// <para>
        /// Implementation in pseudo-LINQ: 
        ///   manager.Entities.Where(e => e.Property == matchValue && e.Key != entity.Key)
        /// </para><para>
        /// The consumer of such a query would likely (a) set it's QueryStrategy to either DataSourceOnly or CacheOnly
        /// and (b) execute it with "Any()" to discover if there are any duplicates.
        /// See <see cref="UniqueVerifier"/> for an example.
        /// </para>
        /// </remarks>
        public static ITypedEntityQuery Build(Entity entity, string valuePropertyName, object matchValue)
        {
            var isDuplicate = 
                PredicateBuilder.And(
                    PredicateBuilder.Make(valuePropertyName, FilterOperator.IsEqualTo, matchValue));

            var keyProps = entity.EntityAspect.EntityMetadata.KeyProperties;

            foreach (var keyProperty in keyProps)
            {
                isDuplicate = isDuplicate.And(
                    PredicateBuilder.Make(keyProperty.Name, FilterOperator.IsNotEqualTo, keyProperty.GetValue(entity))
                    );
            }

            var query = EntityQuery.Create(entity.GetType(), entity.EntityAspect.EntityManager);
            return query.Where(isDuplicate);

        }
    }
}