﻿using System;
using IdeaBlade.EntityModel;
using IdeaBlade.Validation;
// ReSharper disable RedundantThisQualifier
namespace Model
{
    /// <summary>
    /// Validate that no other entity of this type has the same property value.
    /// </summary>
    /// <remarks>
    /// This validation attempts to determine if the proposed property value is in use by any other
    /// entity on the server.
    /// For example, it could check to see if the proposed Customer.Name, "Acme", is the name of
    /// any other Customer entity. 
    /// In pseudo-LINQ terms, the validation fails if 
    /// var testCustId = testCustomer.Id;
    /// manager.Customers.Where(c => c.Name == "Acme" && c.Id != testCustomerId).Any()
    /// ADDED: test that no other entity of this type in cache has this same property value
    /// ADDED: prevent property from being updated with invalid value (IS THIS DESIREABLE IN YOUR APP?)
    /// <para>
    /// While an interesting verifier example, this one is not a good idea;
    /// verifiers should only depend on information immediately available;
    /// this verifier requires a database query.
    /// </para>
    /// <para>
    /// RACE CONDITIONS MEAN THAT THIS VALIDATION MAY NOT PREVENT DUPLICATES.
    /// Someone could insert a duplicate entity between the time the validation is run 
    /// and when the duplicate entity database save actually occurs.
    /// Only a uniqueness test in the database itself can prevent duplicates with certainty.
    /// Be prepared for a database exception when a saved entity violates uniqueness constraints.
    /// </para><para>
    /// WILL NOT WORK IN SILVERLIGHT!
    /// The test for uniqueness involves a synchronous query which cannot be performed in a Silverlight client.
    /// </para>
    /// </remarks>
    public class UniqueVerifier : PropertyValueVerifier
    {
        public UniqueVerifier(Type entityType, String propertyName, bool isRequired = false, String displayName = null, bool? shouldTreatEmptyStringAsNull = null)
            :base(new PropertyValueVerifierArgs(entityType, propertyName, isRequired, displayName, shouldTreatEmptyStringAsNull))
        {
            // validate when instance is verified and before setting the value in the object
            VerifierOptions.ExecutionModes = VerifierExecutionModes.InstanceAndOnBeforeSetTriggers; 

            VerifierOptions.ShouldExitOnBeforeSetError = true; // want current value to be preserved if fails validation

            IsDuplicateTestWhenDbUnreachable = DefaultIsDuplicateTestWhenDbUnreachable;
        }

        protected override void OnVerifierEngineAttach()
        {
            this.ErrorMessageInfo.ErrorMessage = String.Format("{0} must be unique.", this.DisplayName);
        }

        protected override VerifierResult VerifyValue
        (
                object itemToVerify,
                object valueToVerify,
                TriggerContext triggerContext,
                VerifierContext verifierContext
        )
        {
            var entity = (Entity)itemToVerify;
            var isDuplicate = IsDuplicateQueryBuilder.Build(entity, PropertyName, valueToVerify);

            // Check the cache first - if you comment this out, you can see a test fail!
            if (isDuplicate
                 .With(QueryStrategy.CacheOnly)
                 .Any()
                )
                // Break on "return ..." to see when duplicate is in cache
                return new VerifierResult(ok: false);

            // No duplicate in cache; check the database
            try
            {
                var isUnique = !isDuplicate
                    .With(QueryStrategy.DataSourceOnly)
                    .Any();

                return new VerifierResult(isUnique);
            }
            catch (Exception ex)
            {            
              return IsDuplicateTestWhenDbUnreachable (ex, isDuplicate);
            }

        }

        public Func<Exception, ITypedEntityQuery, VerifierResult> IsDuplicateTestWhenDbUnreachable { get; set; }

        private VerifierResult DefaultIsDuplicateTestWhenDbUnreachable(Exception e, ITypedEntityQuery query)
        {
            if (!query.EntityManager.IsConnected)
            {
                return new VerifierResult(ok: true); // assume is ok to proceed if offline
            }
            throw new InvalidOperationException("Can't check database for duplicates.", e);
        }
    }

    [AttributeUsage(AttributeTargets.Property | AttributeTargets.Field, Inherited = false)]
    public class UniqueAttribute : PropertyValueVerifierAttribute
    {
        protected override Verifier BuildVerifierCore(Type pType, String propertyName)
        {
            return new UniqueVerifier(pType, propertyName, IsRequired, DisplayName, GetShouldTreatEmptyStringAsNull());
        }
    }
}