﻿using System;
using System.Resources;
using System.Globalization;
using System.ComponentModel;
using IdeaBlade.Core;

// TODO: Either eliminate the need for this class or move to IdeaBlade.Extensions project

namespace IdeaBlade.Extensions
{
    public class DevForceResourceManagerEmulator
    {

        private ResourceManager _resourceMan;
        private ResourceManager _fallbackResourceMan;

        private static DevForceResourceManagerEmulator _instance;

        /// <summary>
        /// Get the singleton instance of the emulator
        /// </summary>
        public static DevForceResourceManagerEmulator Instance
        {
            get { return _instance ?? (_instance = new DevForceResourceManagerEmulator()); }
            protected set { _instance = value; } // testing hook
        }

        /// <summary>
        /// Get and set the cached <see cref="System.Resources.ResourceManager"/> instance used by this class.
        /// </summary>
        [EditorBrowsableAttribute(EditorBrowsableState.Advanced)]
        public ResourceManager ResourceManager
        {
            get
            {
                return _resourceMan ?? (_resourceMan = DevForceValidationErrorsResourceManager);
            }
            set
            {
                _resourceMan = value;
            }
        }

        /// <summary>
        /// Get the <see cref="System.Resources.ResourceManager"/> 
        /// currently defined for DevForce validation.
        /// </summary>
        private static ResourceManager DevForceValidationErrorsResourceManager
        {
            get { return new IdeaBlade.Validation.VerifierEngine().ErrorsResourceManager; }
        }


        /// <summary>
        /// Get and set a specific <see cref="CultureInfo"/>
        /// for all resource lookups performed by this class.
        /// If null, lookups use the current thread's CurrentUICulture property. 
        /// </summary>
        [EditorBrowsableAttribute(EditorBrowsableState.Advanced)]
        public CultureInfo Culture { get; set; }


        /// <summary>
        /// Get the resource string for the <see cref="resourceKey"/> name,
        /// looking for it first in a <see cref="System.Resources.ResourceManager"/> in the
        /// assembly which defines the <see cref="resourceType"/>.
        /// If that path does not return a matching resource string,
        /// defer to the <see cref="GetString(string)"/> method.
        /// </summary>
        /// <remarks>
        /// The ResourceManager infers the assembly, the base name, and a namespace for .resources files from the Type. 
        /// The ResourceManager assumes you will be using satellite assemblies and want to use the default ResourceSet class. 
        /// Given a type such as MyCompany.MyProduct.MyType, the ResourceManager will look for a .resources file 
        /// (in the main assembly and satellite assemblies) named "MyCompany.MyProduct.MyType.[culture name.]resources" in the assembly 
        /// that defines MyType.
        /// </remarks>
        public string GetString(Type resourceType, String resourceKey)
        {
            String msg = null;
            if (resourceType != null && !String.IsNullOrWhiteSpace(resourceKey))
            {
                var resourceMan = new ResourceManager(resourceType);
                msg = resourceMan.GetString(resourceKey, Culture);
            }
            return msg ?? GetString(resourceKey);
        }

        /// <summary>
        /// Get the resource string for the <see cref="resourceKey"/> name
        /// from the <see cref="ResourceManager"/> or, failing that,
        /// from the <see cref="FallbackResourceManager"/>.
        /// </summary>
        public string GetString(String resourceKey)
        {
            try
            {
                var msg = ResourceManager.GetString(resourceKey, Culture);
                if (String.IsNullOrEmpty(msg) && ResourceManager != FallbackResourceManager)
                {
                    return FallbackResourceManager.GetString(resourceKey, Culture);
                }
                return msg;
            }
            catch (Exception e)
            {    
                throw new IdeaBladeException(
                    "Failed trying to retrieve resource with key=" + resourceKey, e);
            }

        }

        /// <summary>
        /// Get the <see cref="System.Resources.ResourceManager"/> DevForce uses as a fallback
        /// when the <see cref="ResourceManager"/> fails to return a resource for a resource key.
        /// See <see cref="GetString(string)"/>.
        /// </summary>
        private ResourceManager FallbackResourceManager
        {
            get
            {
                return _fallbackResourceMan ??
                       (_fallbackResourceMan =
                        new ResourceManager(
                            "IdeaBlade.Validation.VerifierErrorMessageResource",
                            typeof(IdeaBlade.Validation.VerifierErrorMessageManager).Assembly));

            }
        }
    }

}
