﻿using System;
using System.Diagnostics;
using System.Linq;
using IdeaBlade.Core;
using IdeaBlade.EntityModel;
using IdeaBlade.Validation;

namespace ValidationLab
{
    /// <summary>
    /// Manages client-side validation of entities before they are saved
    /// by an <see cref="EntityManager"/>.
    /// </summary>
    public class OnSaveValidator
    {

        /// <summary>
        /// Add a <see cref="SavingHandler"/> to the <see cref="manager"/>'s
        /// <see cref="EntityManager.Saving"/> event.
        /// </summary>
        public OnSaveValidator AddSavingHandlerTo<T>(T manager) where T : EntityManager
        {
            manager.Saving -= SavingHandler; // remove previous handler (if any)
            manager.Saving += SavingHandler;
            return this;
        }

        /// <summary>
        /// Remove all handlers that belong to this object
        /// that might have been attached to the <see cref="manager"/>.
        /// </summary>
        public OnSaveValidator RemoveHandlersFrom<T>(T manager) where T : EntityManager
        {
            manager.Saving -= SavingHandler;
            return this;
        }

        /// <summary>
        /// An <see cref="EntityManager.Saving"/> event handler that
        /// calls DevForce instance validation on every entity-to-be-saved.
        /// If there are any validation errors, cancels the save
        /// and passes the errors to the <see cref="OnValidationErrorAction"/>.
        /// </summary>
        /// <remarks>
        /// This simple implementation just calls DevForce validation, invoking
        /// the active instance and property validation rules associated with the entities.
        /// More sophisticated handlers might look for and call custom entity validation
        /// methods and manipulate the entities-to-save list as might be useful
        /// for cross-object and aggregate root entity validations.
        /// </remarks>
        protected virtual void SavingHandler(object sender, EntitySavingEventArgs e)
        {
            var manager = sender as EntityManager;
            Debug.Assert(manager != null, "sender is not an EntityManager");
            var engine = manager.VerifierEngine;

            var errors = new VerifierResultCollection();

            var entitiesToSave = e.Entities.OfType<IEntity>();

            // Validate every entity-to-save, gathering validation errors along the way
            foreach (var entity in entitiesToSave)
            {
                engine.Execute(entity).Where(vr => vr.IsError).ForEach(errors.Add);
            }

            if (0 == errors.Count)
            {
                OnValidationSuccessAction();
                return;
            }

            // Errors - cancel the save and take action
            e.Cancel = true;
            OnValidationErrorAction(errors);

        }

        #region OnValidationError

        /// <summary>
        /// Set the <see cref="Action"/> that should handle validation errors
        /// </summary>
        /// <param name="action">
        /// Action subroutine that takes a <see cref="VerifierResultCollection"/> of
        /// errant <see cref="VerifierResult"/>s.
        /// </param>
        /// <remarks>
        /// If none specified, uses the <see cref="OnValidationErrorDefault"/>
        /// which throws an <see cref="InvalidOperationException"/>.
        /// </remarks>
        public OnSaveValidator OnValidationError(Action<VerifierResultCollection> action)
        {
            OnValidationErrorAction = action;
            return this;
        }

        /// <summary>
        /// Get and set the action to perform if there are validation errors
        /// </summary>
        /// <remarks>
        /// Default action is <see cref="OnValidationErrorDefault"/>
        /// </remarks>
        private Action<VerifierResultCollection> OnValidationErrorAction
        {
            get { return _onValidationErrorAction ?? OnValidationErrorDefault; }
            set { _onValidationErrorAction = value; }
        }

        private Action<VerifierResultCollection> _onValidationErrorAction;

        /// <summary>Default <see cref="OnValidationError"/>.</summary>
        /// <exception cref="InvalidOperationException">Default behavior</exception>
        protected virtual void OnValidationErrorDefault(VerifierResultCollection errors)
        {
            throw new InvalidOperationException(
                String.Format(
                    "Cannot save due to {0} validation error(s) including '{1}'",
                    errors.Count, errors.First().Message
                    ));
        }

        #endregion

        #region OnValidation Success

        // Useful when testing OnSaveValidator but no known value in production code
        // Harmless in production code.

        /// <summary>
        /// Set the <see cref="Action"/> that should handle validation errors
        /// </summary>
        /// <param name="action">
        /// Action subroutine that takes a <see cref="VerifierResultCollection"/> of
        /// errant <see cref="VerifierResult"/>s.
        /// </param>
        /// <remarks>Does nothign if no action specified.</remarks>
        public OnSaveValidator OnValidationSuccess(Action action)
        {
            OnValidationSuccessAction = action;
            return this;
        }

        /// <summary>
        /// Get and set the action to perform if there are NO validation errors.
        /// </summary>
        /// <remarks>
        /// Default action is to do nothing
        /// </remarks>
        private Action OnValidationSuccessAction
        {
            get { return _onValidationSuccessAction ?? delegate { }; }
            set { _onValidationSuccessAction = value; }
        }

        private Action _onValidationSuccessAction;

        #endregion

    }
}