﻿using System;
using System.Linq;
using IdeaBlade.EntityModel;
using SoftwareApproach.TestingExtensions;
using Model;
using Microsoft.VisualStudio.TestTools.UnitTesting;

// ReSharper disable InconsistentNaming
// ReSharper disable LocalizableElement

// TESTS FOR .NET ONLY - WON'T WORK IN SILVERLIGHT
// See Model.UniqueVerifier
namespace ValidationLab
{

    [TestClass]
    public class When_Customer_name_must_be_unique_using_IsDuplicateQueryBuilder
        : When_Customer_name_must_be_unique
    {

        /// <summary>Test initialization</summary>
        protected override void SetContext()
        {
            GetAlfreds();
        }

        [TestMethod]
        public void Alfreds_can_be_name_of_entity_with_AlfredsId()
        {
            IsCustomerNameUnique(Alfreds, AlfredsName).ShouldBeTrue();
        }

        [TestMethod]
        public void Alfreds_cannot_be_renamed_to_Simons()
        {
            GetSimons();

            // remove Simons from cache so we test uniqueness in db
            Manager.RemoveEntity(Simons);

            IsCustomerNameUnique(Alfreds, SimonsName).ShouldBeFalse();
        }


        [TestMethod]
        public void Alfreds_cannot_be_name_of_different_cached_entity()
        {
            // Add a new Customer to cache with Alfreds name
            // This sets up a "duplicate name" conflict
            var conflictCustomer = new Customer { CompanyName = AlfredsName };
            Manager.AddEntity(conflictCustomer);
            // The name of the "real" Alfreds Customer is no longer unique
            // because we created a new entity with its name and added that to cache

            // Test would fail verifier didn't check the cache as well as the db
            IsCustomerNameUnique(Alfreds, AlfredsName).ShouldBeFalse();
        }


        /// <summary>
        /// For smoke testing <see cref="IsDuplicateQueryBuilder"/>.
        /// If the proposed CompanyName is in use for a Customer with a different Id
        /// </summary>
        /// <param name="cust">Customer which would receive this name.</param>
        /// <param name="proposedName">Proposed Customer.CompanyName.</param>
        /// <returns>
        /// True if a Customer entity with this Id can receive this name; 
        /// false if a different Customer entity already has this name.
        /// </returns>
        private bool IsCustomerNameUnique(Customer cust, string proposedName)
        {
            var isDuplicate = IsDuplicateQueryBuilder.Build(cust, "CompanyName", proposedName);

            // Check the cache first
            if (isDuplicate
                .With(QueryStrategy.CacheOnly)
                .Any()
                )
                // Break on "return false" to see when duplicate is in cache
                return false;


            // No duplicate in cache; check the database
            return !isDuplicate
                        .With(QueryStrategy.DataSourceOnly)
                        .Any();
        }

    }

    [TestClass]
    public class When_Customer_name_must_be_unique_using_UniqueVerifier
        : When_Customer_name_must_be_unique
    {
        // A Unique attribute was added to Customer.CompanyName in Customer.cs
        /// <summary>Test initialization</summary>
        protected override void SetContext()
        {
            GetSimons();
        }

        [TestMethod]
        public void Simons_name_can_be_changed_to_Simons()
        {
            Simons.CompanyName = "Some crazy name";
            SimonsHasErrors.ShouldBeFalse(
                "Simons.CompanyName has errors when it set to 'crazy'. First error is '{0}'",
                SimonsFirstErrorMessage);

            Simons.CompanyName = SimonsName; // the name in the db

            SimonsHasErrors.ShouldBeFalse(
                "Simons.CompanyName has errors when its name is restored. First error is '{0}'",
                SimonsFirstErrorMessage);
        }

        [TestMethod]
        public void Simons_name_cannot_be_changed_to_Alfreds()
        {
            GetAlfreds();
            // remove Alfreds from cache so we test uniqueness in db
            Manager.RemoveEntity(Alfreds);

            Simons.CompanyName = AlfredsName; // a name in the db

            SimonsHasErrors.ShouldBeTrue(
                "Simons name can be set to Alfreds w/o validation error; it is '{0}'.",
                Simons.CompanyName);

            SimonsFirstErrorMessage.ShouldEqual(
                "Company Name must be unique.",
                "Unexpected error message '{0}'",
                SimonsFirstErrorMessage);

            Simons.CompanyName.ShouldEqual(SimonsName,
                "Simons pending CompanyName should be unchanged");
        }

        [TestMethod]
        public void Simons_cannot_be_name_of_different_cached_entity()
        {
            const string newCustomerName = "<new customer>";

            // Add a new Customer to cache 
            var conflictCustomer = new Customer { CompanyName = newCustomerName };
            Manager.AddEntity(conflictCustomer);

            Simons.CompanyName = newCustomerName; // the name of entity in cache

            // Test would fail if verifier didn't check the cache as well as the db
            SimonsHasErrors.ShouldBeTrue(
                "Simons name can be set to Alfreds w/o validation error; it is '{0}'.",
                Simons.CompanyName);

            SimonsFirstErrorMessage.ShouldEqual(
                "Company Name must be unique.",
                "Unexpected error message '{0}'",
                SimonsFirstErrorMessage);

            Simons.CompanyName.ShouldEqual(SimonsName, 
                "Simons pending CompanyName should be unchanged");
        }

        private bool SimonsHasErrors
        {
            get { return Simons.EntityAspect.ValidationErrors.Any(); }
        }

        private string SimonsFirstErrorMessage
        {
            get
            {
                return SimonsHasErrors
                    ? Simons.EntityAspect.ValidationErrors.First().Message
                    : String.Empty;
            }
        }

    }

    [TestClass]
    public class When_Customer_name_must_be_unique
    {
        [TestInitialize]
        public void TestInitialize()
        {
            Manager = new NorthwindIBEntities();
            SetContext();
        }

        /// <summary>
        /// Override to set sub-class test initialization context.
        /// </summary>
        protected virtual void SetContext() { }

        protected NorthwindIBEntities Manager { get; set; }

        protected void GetAlfreds()
        {
            Alfreds = Manager.Customers.First(c => c.CustomerID == AlfredsId);
            AlfredsName = Alfreds.CompanyName;
        }

        protected void GetSimons()
        {
            Simons = Manager.Customers.First(c => c.CustomerID == SimonsId);
            SimonsName = Simons.CompanyName;
        }


        public Customer Alfreds { get; private set; }
        public string AlfredsName { get; set; }
        public Customer Simons { get; private set; }
        public string SimonsName { get; set; }
        public static Guid AlfredsId = new Guid("785efa04-cbf2-4dd7-a7de-083ee17b6ad2");
        public static Guid SimonsId = new Guid("657d32bc-158c-4d11-ac49-18981ce1e205");

    }

}