﻿using System;
using System.Linq;
using IdeaBlade.EntityModel;
using SoftwareApproach.TestingExtensions;
using Model;
using Microsoft.VisualStudio.TestTools.UnitTesting;

// ReSharper disable InconsistentNaming
// ReSharper disable LocalizableElement;
// ReSharper disable RedundantArgumentDefaultValue
// ReSharper disable PossibleNullReferenceException

namespace ValidationLab
{
    [TestClass]
    public class When_client_saves
    {
        [TestInitialize]
        public void TestInitialize()
        {
            savingHandlerCalled = false;
            Manager = new NorthwindIBEntities();
            Manager.Saving += TestSavingHandler;
            OnSaveValidator = new OnSaveValidator();
        }

        [TestMethod]
        public void no_problem_when_nothing_to_save()
        {
            SaveChanges();
        }

        [TestMethod]
        public void good_Customer_passes_validation()
        {
            var cust = CreateModifiedTestCustomer();
            cust.CompanyName = "SomeValidName"; // required
            cust.City = "Smallville"; // required
            cust.Country = "USA"; // required

            OnSaveValidator.AddSavingHandlerTo(Manager);

            SaveChanges();
        }

        [TestMethod]
        public void multiple_good_Customers_pass_validation()
        {
            CreateModifiedTestCustomer("SomeValidName", "New York City", "USA");
            CreateModifiedTestCustomer("AnotherValidName", "Nairobi", "Kenya");

            OnSaveValidator.AddSavingHandlerTo(Manager);

            SaveChanges();
        }

        [TestMethod]
        public void Customer_without_Country_fails_validation()
        {
            // Country is required ... should fail without it
            CreateModifiedTestCustomer("SomeValidName", city: "Smallville", country: null);

            OnSaveValidator
                .AddSavingHandlerTo(Manager)
                .OnValidationSuccess(() => Assert.Fail("Expected a validation error."))
                .OnValidationError(
                     errors =>
                     {
                         var err = errors.FirstOrDefault(_ => _.PropertyNames.Contains("Country"));
                         err.ShouldNotBeNull("Expected validation error for Country");
                         err.Message.ShouldEqual(
                             "Country is required", 
                             "Unexpected validation error: '{0}'", err.Message);
                     }
                 );

            SaveChanges();
        }

        [TestMethod]
        public void multiple_bad_Customers_fail_validation()
        {
            var cust1 = CreateModifiedTestCustomer("SomeValidName", country: String.Empty);
            var cust2 = CreateModifiedTestCustomer("ThisCompanyNameExceedsTheMaximumLengthByABitTooMuch", country: "Kenya");

            OnSaveValidator
                .AddSavingHandlerTo(Manager)
                .OnValidationSuccess(() => Assert.Fail("Expected validation errors."))
                .OnValidationError(
                     errors =>
                     {
                         var err =
                             errors.FirstOrDefault(_ => _.TargetInstance == cust1 && _.PropertyNames.Contains("Country"));
                         err.ShouldNotBeNull("Expected validation error for cust1.Country");
                         err.Message.ShouldEqual(
                             "Country is required",
                             "Unexpected cust1 validation error: '{0}'", err.Message);

                         err = errors.FirstOrDefault(_ => _.TargetInstance == cust2 && _.PropertyNames.Contains("CompanyName"));
                         err.ShouldNotBeNull("Expected validation error for cust2.CompanyName");
                         err.Message.Contains("CompanyName cannot be longer than")
                             .ShouldBeTrue("Unexpected cust2 validation error: '{0}'", err.Message);
                     }
                 );
            SaveChanges();
        }

        #region Helpers

        private NorthwindIBEntities Manager { get; set; }
        private OnSaveValidator OnSaveValidator { get; set; }

        private Customer CreateModifiedTestCustomer(
            string companyName = null,
            string city = null,
            string country = null)
        {
            var cust = new Customer {CompanyName = companyName, City=city, Country = country};
            Manager.AttachEntity(cust, EntityState.Modified);
            return cust;
        }

        private void TestSavingHandler(object sender, EntitySavingEventArgs e)
        {
            savingHandlerCalled = true;
            e.Cancel = true; // Don't actually save in these offline tests
        }


        private void SaveChanges()
        {
            try
            {
                Manager.SaveChanges(); // what you would do
            }
                // What we must do to keep DevForce from obscuring an Assert failure
                // ignore this nonsense in your production code
            catch (Exception ex)
            {
                var innerException = ex.InnerException;
                // throw inner exception if its an assertion failure
                if (null != innerException && innerException.GetType() == typeof (AssertFailedException))
                {
                    throw innerException;
                }
                throw;
            }
            AssertSavingEventRaised();
        }

        private void AssertSavingEventRaised()
        {
            savingHandlerCalled.ShouldBeTrue("Saving event was not raised");
        }

        private bool savingHandlerCalled;

        #endregion

    }
}
