﻿using System;
using System.ComponentModel;
using System.Linq;
using IdeaBlade.Core.ComponentModel;
using IdeaBlade.EntityModel;
using IdeaBlade.Validation;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Model;

// ReSharper disable InconsistentNaming
// ReSharper disable RedundantDefaultFieldInitializer
// ReSharper disable RedundantArgumentName

namespace ValidationLab
{

    [TestClass]
    public class When_validating_Customer_in_cache
    {
        [TestInitialize]
        public void TestInitialize()
        {
            manager = new EntityManager(shouldConnect: false);
            testCustomer = new Customer
            {
                CompanyName = "Acme",
                Country = "Freedonia",
            };
            manager.AddEntity(testCustomer);
        }

        [TestMethod]
        public void testCustomer_is_error_free_at_start()
        {
            Assert.IsFalse(testCustIndei.HasErrors, 
                "Should be error free at start");
        }


        /// <summary>The <see cref="StringLengthVerifier"/>
        /// adorning the <see cref="Customer.CompanyName"/> property
        /// reports a validation error when the CompanyName is too long.</summary>
        /// <remarks>
        /// DevForce code generation added the DevForce stock <see cref="StringLengthVerifier"/>
        /// to the <see cref="Customer.CompanyName"/> property using the following attribute,
        /// [IbVal.StringLengthVerifier(MaxValue=40, IsRequired=true, ErrorMessageResourceName="Customer_CompanyName")]
        /// This test shows that a validation error results when setting that property
        /// to a name of more than 40 characters.
        /// </remarks>
        [TestMethod]
        public void has_error_when_CompanyName_is_too_long()
        {
            testCustomer.CompanyName = "The quick brown fox jumped over the lazy dog";

            const string expected = "CompanyName cannot be longer than";
            var actual = FirstErrorMessage;
            Assert.IsTrue(actual.StartsWith(expected),
                String.Format("Expected message starting '{0}'; actual message: '{1}'", expected, actual));

            // A better assert when we know the exact message:
            // Assert.AreEqual("CompanyName cannot be longer than 40 character(s),
            //                 "GetFirstErrorMessage);
        }

        /// <summary>The <see cref="StringLengthVerifier"/>
        /// adorning the <see cref="Customer.CompanyName"/> property
        /// reports a validation error when the CompanyName is too long.</summary>
        /// <remarks>
        /// See the <see cref="StringLengthVerifierAttribute"/> that DevForce code generation added to  
        /// the <see cref="Customer.CompanyName"/> property.
        /// Note the "IsRequired=true" parameter which means that the CompanyName must also have a value
        /// in addition to being less than 40 characters.
        /// By default, DevForce treats <see cref="String.Empty"/> as a missing value
        /// as it would a null value.
        /// </remarks>
        [TestMethod]
        public void has_error_when_CompanyName_is_empty_string()
        {
            testCustomer.CompanyName = String.Empty;
            Assert.AreEqual( "CompanyName is required", FirstErrorMessage);
        }

        /// <summary>Demonstrates the custom <see cref="NotEmptyGuidVerifier"/></summary>
        /// <remarks>
        /// Get an error because of custom validation rule
        /// <see cref="NotEmptyGuidVerifier"/> added in
        /// <see cref="ModelVerifierProvider"/>.
        /// Also demonstrates the use of <see cref="INotifyDataErrorInfo"/>
        /// to detect validation errors just as a UI control might do.
        /// </remarks>
        [TestMethod]
        public void has_error_when_set_CustomerID_empty()
        {
            testCustomer.CustomerID = Guid.Empty;
            // Demonstrate use of INotifyDataErrorInfo interface
            Assert.IsTrue(testCustIndei.HasErrors,"Should have error when ID is Guid.Empty");
            Assert.AreEqual("Id cannot be empty", FirstErrorMessage);
        }

        /// <summary>Test verifier attribute added to Country in partial class</summary>
        [TestMethod]
        public void has_error_when_set_Country_to_empty_string()
        {
            testCustomer.Country = String.Empty;
            // testCustIndei demonstrates use of the INotifyDataErrorInfo interface
            Assert.IsTrue(testCustIndei.HasErrors,"Should have error when Country is String.Empty");
            Assert.AreEqual("Country is required", FirstErrorMessage);

        }

        [TestMethod]
        public void does_not_validate_detached_Customer()
        {
            testCustomer.EntityAspect.RemoveFromManager(clearQueryCache: false); // detach the customer
            Assert.IsTrue(testCustomer.EntityAspect.EntityState.IsDetached(),
                "Expected testCustomer removed from manager to be detached");

            // The following would provoke an error in an attached Customer
            // as shown with test "has_error_when_CompanyName_is_empty_string"
            testCustomer.CompanyName = String.Empty;

            // but it does not produce an error for a detached entity
            Assert.IsFalse(testCustIndei.HasErrors,
                String.Format("Deleted customer has validation error: '{0}'", FirstErrorMessage));
        }

        [TestMethod]
        public void still_validates_deleted_Customer()
        {
            testCustomer.EntityAspect.AcceptChanges(); // pretend it is unmodified entity retrieved from database
            testCustomer.EntityAspect.Delete();        // schedule the customer to be deleted upon Save

            // The following would provoke an error in a non-deleted Customer
            // as shown with test "has_error_when_CompanyName_is_empty_string"
            testCustomer.CompanyName = String.Empty;

            // and still does in a deleted entity
            Assert.AreEqual("CompanyName is required", FirstErrorMessage);
        }

        [TestMethod]
        public void does_not_validate_when_new_Customer_is_unattached()
        {

            // a new'ed Customer is detached until added to an EntityManager
            testCustomer = new Customer {CompanyName = "Valid customer name"}; 

            // The following would provoke an error in an attached Customer
            // as shown with test "has_error_when_CompanyName_is_empty_string"
            testCustomer.CompanyName = String.Empty;

            // But it doesn't because the Customer is not yet associated with an EntityManager
            // and therefore there is no VerifierEngine to evaluate the entity's validity
            Assert.IsFalse(testCustIndei.HasErrors,
                String.Format("New unattached customer has validation error: '{0}'", FirstErrorMessage));

            // nor does adding the entity to a manager cause validation
            manager.AddEntity(testCustomer);

            Assert.IsFalse(testCustIndei.HasErrors,
                String.Format("Adding new customer to manager produced validation error: '{0}'", FirstErrorMessage));
        }

        /// <summary>Demonstrates that standard .NET validations work.</summary>
        /// <remarks>
        /// Get an error when violate City validation attributes from
        /// <see cref="System.ComponentModel.DataAnnotations"/>
        /// </remarks>
        [Ignore] // ToDo: find out why mixing .NET and DF validation rules is not working
        [TestMethod]
        public void has_error_when_set_City_empty()
        {
            testCustomer.City = "Valid city";
            Assert.IsFalse(testCustIndei.HasErrors, 
                "Should not have error when City given a valid value; message is "+FirstErrorMessage);

            testCustomer.City = null; 
            Assert.IsTrue(testCustIndei.HasErrors, "Should have error when City is null");
            Assert.AreEqual("City is required", FirstErrorMessage);
        }

        private string FirstErrorMessage
        {
            get
            {
                var err = testCustomer.EntityAspect.ValidationErrors.FirstOrDefault();
                return null == err ? String.Empty : err.Message;
           }
        }

        protected EntityManager manager;
        protected Customer testCustomer;
        protected IDataErrorInfo testCustIdei { get { return testCustomer; } }
        protected INotifyDataErrorInfo testCustIndei { get { return testCustomer; } }

    }
}